
from itertools import combinations


def facesiter(simplex):
    for i in range(len(simplex)):
        yield simplex[:i] + simplex[(i + 1):]


def flattening_simplex(simplices):
    for simplex in simplices:
        for point in simplex:
            yield point


def get_allpoints(simplices):
    return set(flattening_simplex(simplices))


def faces(simplices):
    faceset = set()
    for simplex in simplices:
        numnodes = len(simplex)
        for r in range(numnodes, 0, -1):
            for face in combinations(simplex, r):
                faceset.add(tuple(sorted(face)))
    return faceset



import unittest
import numpy as np

from mogutda import SimplicialComplex, AlphaComplex


# Simplicial Complex
# ring
ring1d_sc = [(i, i + 1) for i in range(10)] + [(10, 0)]
# cylinder with hole
cylinder_sc = [(0, 1, 2), (1, 2, 3), (2, 3, 0), (3, 0, 1)]
# torus (https://page.mi.fu-berlin.de/rote/Papers/pdf/Computational+topology:+an+introduction.pdf)
torus_sc = [
    (1, 2, 4),
    (4, 2, 5),
    (2, 3, 5),
    (3, 5, 6),
    (5, 6, 1),
    (1, 6, 2),
    (6, 7, 2),
    (7, 3, 2),
    (1, 3, 4),
    (3, 4, 6),
    (4, 6, 7),
    (4, 5, 7),
    (5, 7, 1),
    (7, 3, 1),
]
# sphere
twosphere_c = [(1, 3, 4), (1, 2, 3), (2, 3, 4), (1, 2, 4)]

# Alpha Complex
# circular ring
ring = np.array(
    [
        np.random.normal(loc=1, scale=0.01) * np.array([np.cos(t), np.sin(t)])
        for t in np.linspace(0, 2 * np.pi, 101)
    ]
)[1:]
# spherical ball
sphere = np.array(
    [
        [np.sin(theta) * np.cos(phi), np.sin(theta) * np.sin(phi), np.cos(theta)]
        for theta in np.linspace(0.02, np.pi, 51)
        for phi in np.linspace(0.02, 2 * np.pi, 51)
    ]
)


class test_simcomplex(unittest.TestCase):
    def setUp(self):
        self.sc1 = SimplicialComplex(simplices=[(1, 2), (2, 3), (1, 2, 3)])
        self.sc2 = AlphaComplex(ring, 0.05)
        self.ring1d = SimplicialComplex(simplices=ring1d_sc)
        self.cylinder = SimplicialComplex(simplices=cylinder_sc)
        self.torus = SimplicialComplex(simplices=torus_sc)
        self.twosphere = SimplicialComplex(simplices=twosphere_c)

    def tearDown(self):
        pass

    def test_ring1d(self):
        self.assertEqual(self.ring1d.betti_number(0), 1)
        self.assertEqual(self.ring1d.betti_number(1), 1)
        self.assertEqual(self.ring1d.betti_number(2), 0)

    def test_cylinder(self):
        self.assertEqual(self.cylinder.betti_number(0), 1)
        self.assertEqual(self.cylinder.betti_number(1), 0)
        self.assertEqual(self.cylinder.betti_number(2), 1)
        self.assertEqual(self.cylinder.euler_characteristics(), 2)

    def test_torus(self):
        self.assertEqual(self.torus.betti_number(0), 1)
        self.assertEqual(self.torus.betti_number(1), 2)
        self.assertEqual(self.torus.betti_number(2), 1)
        self.assertEqual(self.torus.betti_number(3), 0)
        self.assertEqual(self.torus.euler_characteristics(), 0)
        self.assertEqual(len(self.torus.generate_graph()[0]), 7)
        self.assertEqual(len(self.torus.generate_graph()[1]), 21)

    def test_twosphere(self):
        self.assertEqual(self.twosphere.betti_number(0), 1)
        self.assertEqual(self.twosphere.betti_number(1), 0)
        self.assertEqual(self.twosphere.betti_number(2), 1)
        self.assertEqual(self.twosphere.euler_characteristics(), 2)

    def test_sc1(self):
        self.assertEqual(self.sc1.betti_number(0), 1)
        self.assertEqual(self.sc1.betti_number(1), 0)
        self.assertEqual(self.sc1.betti_number(2), 0)
        self.assertEqual(self.sc1.euler_characteristics(), 1)

    def test_sc2(self):
        self.assertEqual(self.sc2.betti_number(0), 1)
        self.assertEqual(self.sc2.betti_number(1), 1)
        self.assertEqual(self.sc2.betti_number(2), 0)
        self.assertEqual(self.sc2.euler_characteristics(), 0)

    # def test_sc2_sparse(self):
    #     self.assertEqual(self.sc2.betti_number(0, eps=0.001), 1)
    #     self.assertEqual(self.sc2.betti_number(1, eps=0.001), 1)
    #     self.assertEqual(self.sc2.betti_number(2, eps=0.001), 0)


if __name__ == "__main__":
    unittest.main()




from itertools import product

import networkx as nx
from scipy.spatial import distance

from .abssimcomplex import SimplicialComplex


class VietorisRipsComplex(SimplicialComplex):
    def __init__(
            self,
            points,
            epsilon,
            labels=None,
            distfcn=distance.euclidean
    ):
        super(VietorisRipsComplex, self).__init__()
        self.pts = points
        self.labels = (range(len(self.pts))
                       if labels is None or len(labels) != len(self.pts)
                       else labels)
        self.epsilon = epsilon
        self.distfcn = distfcn
        self.network = self.construct_network(
            self.pts, self.labels, self.epsilon, self.distfcn
        )
        self.import_simplices(map(tuple, nx.find_cliques(self.network)))

    def construct_network(self, points, labels, epsilon, distfcn):
        g = nx.Graph()
        g.add_nodes_from(labels)
        zips = zip(points, labels)
        for pair in product(zips, zips):
            if pair[0][1] != pair[1][1]:
                dist = distfcn(pair[0][0], pair[1][0])
                if dist < epsilon:
                    g.add_edge(pair[0][1], pair[1][1])
        return g




from scipy.spatial import Delaunay, distance

from .abssimcomplex import SimplicialComplex
from .utils import facesiter, get_allpoints


def calculate_distmatrix(points, labels, distfcn):
    return {
        (labels[i], labels[j]): distfcn(points[i], points[j])
        for i in range(len(labels)) for j in range(len(labels))
    }


def contain_detachededges(simplex, distdict, epsilon):
    if len(simplex) == 2:
        return distdict[simplex[0], simplex[1]] > 2*epsilon
    else:
        for face in facesiter(simplex):
            contained = contain_detachededges(face, distdict, epsilon)
            if contained:
                return True
        return False


class AlphaComplex(SimplicialComplex):
    def __init__(
            self,
            points,
            epsilon,
            labels=None,
            distfcn=distance.euclidean
    ):
        super(AlphaComplex, self).__init__()
        self.pts = points
        self.labels = (range(len(self.pts))
                       if (labels is None or len(labels) != len(self.pts))
                       else labels)
        self.epsilon = epsilon
        self.distfcn = distfcn
        self.import_simplices(
            self.construct_simplices(self.pts, self.labels, self.epsilon, self.distfcn)
        )

    def construct_simplices(self, points, labels, epsilon, distfcn):
        distdict = calculate_distmatrix(points, labels, distfcn)

        delaunayobj = Delaunay(points)

        simplices = []
        for simplexnda in delaunayobj.simplices:
            simplex = tuple(simplexnda)

            detached = [contain_detachededges(face, distdict, epsilon) for face in facesiter(simplex)]

            if True in detached and len(simplex) > 2:
                simplices += [
                    face
                    for face, notkeep in zip(facesiter(simplex), detached)
                    if not notkeep
                ]
            else:
                simplices.append(simplex)
        # simplices = map(lambda simplex: tuple(sorted(simplex)), simplices)
        simplices = [tuple(sorted(simplex)) for simplex in simplices]
        simplices = list(set(simplices))

        allpts = get_allpoints(simplices)
        simplices += [(point,) for point in (set(labels)-allpts)]

        return simplices




from itertools import combinations
import warnings

import numpy as np
from scipy.sparse import dok_matrix
from scipy.sparse.linalg import aslinearoperator
from scipy.linalg.interpolative import estimate_rank
import networkx as nx
import matplotlib.pyplot as plt

from .utils import faces


class SimplicialComplex:
    def __init__(self, simplices=[]):
        self.import_simplices(simplices=simplices)

    def import_simplices(self, simplices=[]):
        # self.simplices = list(map(lambda simplex: tuple(sorted(simplex)), simplices))
        self.simplices = [tuple(sorted(simplex)) for simplex in simplices]
        self.face_set = faces(self.simplices)

    def n_faces(self, n):
        # return list(filter(lambda face: len(face) == n + 1, self.face_set))
        return [face for face in self.face_set if len(face) == n+1]

    def boundary_operator(self, i):
        source_simplices = self.n_faces(i)
        target_simplices = self.n_faces(i - 1)

        if len(target_simplices) == 0:
            S = dok_matrix((1, len(source_simplices)), dtype=np.float64)
            S[0, 0:len(source_simplices)] = 1
        else:
            source_simplices_dict = {
                source_simplices[j]: j for j in range(len(source_simplices))
            }
            target_simplices_dict = {
                target_simplices[i]: i for i in range(len(target_simplices))
            }

            S = dok_matrix(
                (len(target_simplices), len(source_simplices)), dtype=np.float64
            )
            for source_simplex in source_simplices:
                for a in range(len(source_simplex)):
                    target_simplex = source_simplex[:a] + source_simplex[(a + 1) :]
                    i = target_simplices_dict[target_simplex]
                    j = source_simplices_dict[source_simplex]
                    S[i, j] = -1 if a % 2 == 1 else 1  # S[i, j] = (-1)**a
        return S

    def betti_number(self, i, eps=None):
        boundop_i = self.boundary_operator(i)
        boundop_ip1 = self.boundary_operator(i + 1)

        if i == 0:
            boundop_i_rank = 0
        else:
            if eps is None:
                try:
                    boundop_i_rank = np.linalg.matrix_rank(boundop_i.toarray())
                except (np.linalg.LinAlgError, ValueError):
                    boundop_i_rank = boundop_i.shape[1]
            else:
                boundop_i_rank = estimate_rank(aslinearoperator(boundop_i), eps)

        if eps is None:
            try:
                boundop_ip1_rank = np.linalg.matrix_rank(boundop_ip1.toarray())
            except (np.linalg.LinAlgError, ValueError):
                boundop_ip1_rank = boundop_ip1.shape[1]
        else:
            boundop_ip1_rank = estimate_rank(aslinearoperator(boundop_ip1), eps)

        return (boundop_i.shape[1] - boundop_i_rank) - boundop_ip1_rank

    def euler_characteristics(self):
        max_n = max(map(len, self.simplices))
        return sum(
            [(-1 if a % 2 == 1 else 1) * self.betti_number(a) for a in range(max_n)]
        )

    def generate_graph(self, simplices=None):
        """ Generate edges and vertices in sets """
        if len(self.simplices) == 0 and simplices is None:
            raise Exception("No simplices to create a graph from. Please specify one.")
        elif simplices is not None:
            self.import_simplices(simplices=simplices)

        vertices = set()
        edges = set()

        for simplex in self.simplices:
            for vertex1, vertex2 in combinations(simplex, 2):
                vertices.add(vertex1)
                vertices.add(vertex2)
                edges.add(tuple(sorted([vertex1, vertex2])))
        return vertices, edges

    def draw(self, vertex_labels=True, edge_labels=False):
        """ Draw the Simplicial Complex """
        vertices, edges = self.generate_graph()

        if len(vertices) + len(edges) > 1000:
            warnings.warn(
                "The number of elements to draw exceeds 100. To still draw the plot, call the function with override_100_limit = True. "
            )
        else:
            G = nx.Graph()
            for vertex in vertices:
                G.add_node(vertex)
            for edge in edges:
                G.add_edge(edge[0], edge[1])
            pos = nx.spring_layout(G)

            nx.draw(G, pos)

            labels = {}
            if vertex_labels is True:
                for vertex in vertices:
                    labels[vertex] = str(vertex)
            nx.draw_networkx_labels(G, pos, labels, font_size=16)

            if edge_labels is True:
                edge_labels = {}
                for edge in edges:
                    edge_labels[edge] = "{}:{}".format(edge[0], edge[1])
                nx.draw_networkx_edge_labels(G, pos, edge_labels, font_size=3)
            plt.plot()




from setuptools import setup

def readme():
    with open('README.md') as f:
        return f.read()


def package_description():
    text = open('README.md', 'r').read()
    startpos = text.find('## Introduction')
    return text[startpos:]


def install_requirements():
    return [package_string.strip() for package_string in open('requirements.txt', 'r')]


setup(name='mogutda',
      version="0.3.2",
      description="Topological Data Analysis in Python",
      long_description=package_description(),
      long_description_content_type='text/markdown',
      classifiers=[
          "Topic :: Scientific/Engineering :: Mathematics",
          "Topic :: Scientific/Engineering :: Physics",
          "Programming Language :: Python :: 3.6",
          "Programming Language :: Python :: 3.7",
          "Programming Language :: Python :: 3.8",
          "Programming Language :: Python :: 3.9",
          "License :: OSI Approved :: MIT License",
          "Intended Audience :: Developers",
          "Intended Audience :: Education",
          "Intended Audience :: Information Technology",
          "Intended Audience :: Science/Research"
      ],
      keywords="mogutda numerics topology data",
      url="https://github.com/stephenhky/MoguTDA",
      author="Kwan-Yuet Ho",
      author_email="stephenhky@yahoo.com.hk",
      license='MIT',
      packages=['mogutda',],
      install_requires=install_requirements(),
      include_package_data=True,
      test_suite="test",
      zip_safe=False)

